"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const env_1 = require("../../env");
const _ = require("lodash");
const camelcaseKeys = require('camelcase-keys');
const TestEvents_1 = require("../Mock/Events/TestEvents");
var EventTypes;
(function (EventTypes) {
    EventTypes.toAscii = (value) => {
        return env_1.web3.toAscii(value).replace(/\u0000/g, '');
    };
    EventTypes.guessType = (value) => {
        if (typeof value === 'number') {
            return 'UInt';
        }
        if (typeof value === 'object') {
            return 'Object';
        }
        if (typeof value === 'string') {
            if (env_1.web3.isAddress(value)) {
                return 'Address';
            }
            return 'Bytes32';
        }
        throw Error('unable to guess type of value: ' + value);
    };
    EventTypes.EsEvent = 'EsEvent';
    EventTypes.EsEventProperty = 'EsEventProperty';
    EventTypes.EsEventPropertySchema = {
        EventIndex: 'BigNumber',
        EventPropertyIndex: 'BigNumber',
        Name: 'Bytes32',
        ValueType: 'Bytes32',
        AddressValue: 'String',
        UIntValue: 'BigNumber',
        Bytes32Value: 'Bytes32'
    };
    EventTypes.EsEventSchema = {
        Id: 'BigNumber',
        Type: 'Bytes32',
        Version: 'Bytes32',
        ValueType: 'Bytes32',
        AddressValue: 'String',
        UIntValue: 'BigNumber',
        Bytes32Value: 'Bytes32',
        TxOrigin: 'String',
        Created: 'BigNumber',
        PropertyCount: 'BigNumber'
    };
    EventTypes.TruffleEventSchema = {
        [EventTypes.EsEvent]: EventTypes.EsEventSchema,
        [EventTypes.EsEventProperty]: EventTypes.EsEventPropertySchema
    };
    EventTypes.getPropFromTruffleEventSchemaType = (schemaPropType, truffleValue) => {
        switch (schemaPropType) {
            case 'Address': return truffleValue.toString();
            case 'String': return truffleValue.toString();
            case 'Bytes32': return EventTypes.toAscii(truffleValue);
            case 'BigNumber': return truffleValue.toNumber();
            case 'UInt': return truffleValue.toNumber();
            default: throw Error(`UNKNWON schemaPropType for truffleValue '${truffleValue}'. Make sure your schema is up to date.`);
        }
    };
    EventTypes.getTransmuteEventPropertyFromEsEventProperty = (prop) => {
        let _obj = {};
        switch (prop.ValueType) {
            case 'Bytes32':
                _obj[prop.Name] = EventTypes.toAscii(prop.Bytes32Value);
                break;
            case 'BigNumber':
                _obj[prop.Name] = prop.UIntValue;
                break;
            case 'Address':
                _obj[prop.Name] = prop.AddressValue;
                break;
        }
        return _obj;
    };
    EventTypes.getEsEventFromEsEventValues = (eventValues) => {
        let evt = {};
        _.keys(EventTypes.EsEventSchema).map((k, i) => {
            evt[k] = eventValues[i];
        });
        return evt;
    };
    EventTypes.getEsEventPropertyFromEsEventPropertyValues = (eventPropValues) => {
        let evt = {};
        _.keys(EventTypes.EsEventPropertySchema).map((k, i) => {
            evt[k] = eventPropValues[i];
        });
        return evt;
    };
    EventTypes.getEsEventFromEsEventWithTruffleTypes = (eventType, solEvent) => {
        let event = {};
        let schema = EventTypes.TruffleEventSchema[eventType];
        _.forIn(solEvent, (value, key) => {
            let prop = EventTypes.getPropFromTruffleEventSchemaType(schema[key], value);
            _.extend(event, {
                [key]: prop
            });
        });
        return event;
    };
    EventTypes.getEsEventPropertyFromEsEventPropertyWithTruffleTypes = (eventType, solEvent) => {
        let event = {};
        let schema = EventTypes.TruffleEventSchema[eventType];
        _.forIn(solEvent, (value, key) => {
            let prop = EventTypes.getPropFromTruffleEventSchemaType(schema[key], value);
            _.extend(event, {
                [key]: prop
            });
        });
        return event;
    };
    EventTypes.getValueType = (es) => {
        switch (es.ValueType) {
            case 'Address': return es.AddressValue;
            case 'UInt': return es.UIntValue;
            case 'Bytes32': return es.Bytes32Value;
            default: throw Error('ValueType invalid, must be [UInt, Address, Bytes32]');
        }
    };
    EventTypes.convertMeta = (esEvent) => {
        let metaKeys = ['Type', 'ValueType', 'PropertyCount', 'AddressValue', 'UIntValue', 'Bytes32Value'];
        let objectWithoutEsMeta = _.omit(esEvent, metaKeys);
        let withProperCase = camelcaseKeys(objectWithoutEsMeta);
        return withProperCase;
    };
    EventTypes.payloadReducer = (state = {}, esEvenProperty) => {
        return Object.assign({}, state, {
            [esEvenProperty.Name]: EventTypes.getValueType(esEvenProperty)
        });
    };
    EventTypes.esEventToTransmuteEvent = (esEvent, esEventProps) => tslib_1.__awaiter(this, void 0, void 0, function* () {
        let payload = {};
        let meta = EventTypes.convertMeta(esEvent);
        if (!esEvent.PropertyCount) {
            payload = EventTypes.getValueType(esEvent);
        }
        else {
            esEventProps.forEach((esEventProp) => {
                payload = EventTypes.payloadReducer(payload, esEventProp);
            });
            payload = EventTypes.unflatten(payload);
        }
        return {
            type: esEvent.Type,
            payload: payload,
            meta: meta
        };
    });
    EventTypes.convertCommandToEsEvent = (transmuteCommand) => {
        let valueType = EventTypes.guessType(transmuteCommand.payload);
        let esEvent;
        switch (valueType) {
            case 'Address':
                esEvent = _.cloneDeep(TestEvents_1.addressValueEsEvent);
                esEvent.AddressValue = transmuteCommand.payload;
                break;
            case 'UInt':
                esEvent = _.cloneDeep(TestEvents_1.uIntValueEsEvent);
                esEvent.UIntValue = transmuteCommand.payload;
                break;
            case 'Bytes32':
                esEvent = _.cloneDeep(TestEvents_1.bytes32ValueEsEvent);
                esEvent.Bytes32Value = transmuteCommand.payload;
                break;
            case 'Object':
                esEvent = _.cloneDeep(TestEvents_1.objectValueEsEvent);
                esEvent.PropertyCount = Object.keys(transmuteCommand.payload).length;
                break;
            default:
                throw Error('Unkown valueType: ' + valueType);
        }
        esEvent.Type = transmuteCommand.type;
        return esEvent;
    };
    EventTypes.convertCommandToEsEventProperties = (esEvent, transmuteCommand) => {
        let payload = transmuteCommand.payload;
        let flatPayload = EventTypes.flatten(payload);
        let payloadKeys = _.keys(flatPayload);
        let payloadVals = _.values(flatPayload);
        let payLoadValueTypes = payloadVals.map(EventTypes.guessType);
        let esEventProps = [];
        payloadKeys.forEach((key, i) => {
            let esEventProp;
            let valueType = payLoadValueTypes[i];
            let value = payloadVals[i];
            let name = payloadKeys[i];
            switch (valueType) {
                case 'Address':
                    esEventProp = _.cloneDeep(TestEvents_1.addressValueEsEventProperty);
                    esEventProp.AddressValue = value;
                    break;
                case 'UInt':
                    esEventProp = _.cloneDeep(TestEvents_1.uIntValueEsEventProperty);
                    esEventProp.UIntValue = value;
                    break;
                case 'Bytes32':
                    esEventProp = _.cloneDeep(TestEvents_1.bytes32ValueEsEventProperty);
                    esEventProp.Bytes32Value = value;
                    break;
                default:
                    throw Error('Unkown valueType: ' + valueType);
            }
            esEventProp.EventIndex = esEvent.Id;
            esEventProp.EventPropertyIndex = i;
            esEventProp.Name = name;
            esEventProps.push(esEventProp);
        });
        return esEventProps;
    };
    EventTypes.flatten = (data) => {
        var result = {};
        function recurse(cur, prop) {
            if (Object(cur) !== cur) {
                result[prop] = cur;
            }
            else if (Array.isArray(cur)) {
                for (var i = 0, l = cur.length; i < l; i++)
                    recurse(cur[i], prop + "[" + i + "]");
                if (l == 0)
                    result[prop] = [];
            }
            else {
                var isEmpty = true;
                for (var p in cur) {
                    isEmpty = false;
                    recurse(cur[p], prop ? prop + "." + p : p);
                }
                if (isEmpty && prop)
                    result[prop] = {};
            }
        }
        recurse(data, "");
        return result;
    };
    EventTypes.unflatten = (data) => {
        if (Object(data) !== data || Array.isArray(data))
            return data;
        var regex = /\.?([^.\[\]]+)|\[(\d+)\]/g, resultholder = {};
        for (var p in data) {
            var cur = resultholder, prop = "", m;
            while (m = regex.exec(p)) {
                cur = cur[prop] || (cur[prop] = (m[2] ? [] : {}));
                prop = m[2] || m[1];
            }
            cur[prop] = data[p];
        }
        return resultholder[""] || resultholder;
    };
    EventTypes.eventsFromTransaction = (tx) => {
        let allEvents = tx.logs.map((log) => {
            return EventTypes.getEsEventFromEsEventWithTruffleTypes(log.event, log.args);
        });
        return allEvents;
    };
    EventTypes.reconstructTransmuteEventsFromTxs = (txs) => {
        let esEventsAndEventProps = _.flatten(txs.map(EventTypes.eventsFromTransaction));
        let esEvents = _.filter(esEventsAndEventProps, (obj) => {
            return obj.Id !== undefined;
        });
        let transmuteEvents = [];
        esEvents.forEach((esEvent) => {
            let esEventProps = _.filter(esEventsAndEventProps, (obj) => {
                return obj.EventIndex === esEvent.Id;
            });
            let transmuteEvent = EventTypes.esEventToTransmuteEvent(esEvent, esEventProps);
            transmuteEvents.push(transmuteEvent);
        });
        return transmuteEvents;
    };
})(EventTypes = exports.EventTypes || (exports.EventTypes = {}));
